// PMC (Paolo's Minigame Compressor)'s compressor routine
// Public domain by Paolo Ferraris (pieffe8@libero.it)

// 2003 Sep 15 - modified for zmac (rjm)

#include<iostream>
#include<stdio.h>
#include<string>
#include<vector>
#include<list>
#include<assert.h>

using namespace std;


// generate the tree, by returning the code for each symbol.

// to find the best tree, we use dynamic programming. We find the
// best tree for each range of symbols [a,b]
// (0<=a<=b<l). We find such tree for increasing distances between a and b.

// If a=b, the tree is just a leaf. If b>a, we find the best tree by
// 1) splitting the interval into two parts
// 2) seeing for which splitting we have the best tree by looking at the
//    best tree for the two sub-intervals.

vector<string> pmccode(const vector<int> & occur) {
  int l=occur.size();
  int z,v;
  int a,b,c;
  // total sum of bits for the tree with elements between a and b
  // 1000000=infinity
  vector<vector<int> > parsum(l,vector<int> (l,1000000));
  // the best way to partition a tree between a and b will be
  // between a and parbest[a][b], and between parbest[a][b]+1 and b
  vector<vector<int> > parbest(l,vector<int> (l,0));
  // number of occurrences between a and b
  vector<vector<int> > parocc(l,vector<int> (l,0));
  for(z=0;z<l;z++) {
    parocc[z][z]=occur[z];
    for(v=z+1;v<l;v++)
      parocc[z][v]=parocc[z][v-1]+occur[v];
  }
  // base case: parsum for [a][a] is 0
  for(a=0;a<l;a++)
    parsum[a][a]=0;
  // now we calculate the others
  for(z=1;z<l;z++)//distance between a and b
    for(a=0,b=z;b<l;a++,b++)
      for(c=a;c<b;c++) {// this divides the interval [a,b]
	                // into [a,c] and [c+1,b]
	// total number of bits for the best tree of [a][b] when splitting
	// the interval [a][b] into [a][c] and [c+1][b]
	v=parsum[a][c]+parsum[c+1][b]+parocc[a][b];
	// memorize this splitting if it is better than the previous ones
	if(v<parsum[a][b]) {
	  parsum[a][b]=v;
	  parbest[a][b]=c;
	}
      }
  // we calculate each code by navigating the tree to the leaf
  vector<string> code(l);   // the output
  for(z=0;z<l;z++) {
    a=0;
    b=l-1;
    while(a!=b) {
      if(parbest[a][b]>=z) { // first half?
	code[z]+='0';
	b=parbest[a][b]; // we update the upper bound
      } else { // second half
	code[z]+='1';
	a=parbest[a][b]+1; // we update the lower bound
      }
    }
  }
  return code;
}

//finds the longest occurrence
void findoccur(const string & s, const int & pos, const int & maxdist, int & len, int & dist) {
  int u,t;
  int z=s.size();
  len=0;
  for(u=1;u<=maxdist && pos-u>=0;u++) {// we go back u positions
    for(t=0;t<18 && pos+t<z;t++) // each byte in the sequence
      if(s[pos+t]!=s[pos-u+t]) break; // if the two bytes are different
                                  // the sequence is over
    if((!len || t>len) && t>2) { // a new or better sequence
                                 // of length >=3 ?
      len=t;
      dist=u;
    }
  }
}

int main(int argc,char** argv) {
  FILE* fin;
  FILE* fout;

  vector<int> occur(256,0);
  vector<int> occur2(256,0);
  vector<int> len(256,0);

  int z,v,u,t;
  int pmc1=0;
  if(argc>1 && string(argv[1])=="-1")
    pmc1=1;
  if(argc!=3+pmc1) {
    cout <<"\nSyntax: pmccompr [-1] origin_file destination_file \n";
    cout <<"\t-1\tPMC1 compression\n\n";
    exit(1);
  }
  fin=fopen(argv[1+pmc1],"rb");
  if(!fin) {
    printf("Error opening the input file `%s\'\n",argv[1+pmc1]);
    return 1;
  }
  fout=fopen(argv[2+pmc1],"w");
  if(!fout) {
    printf("Error opening the output file `%s\'\n",argv[2+pmc1]);
    return 1;
  }
  string input;
  // read the input in the string input
  // taking the occurrences of each symbol in occur
  while((z=getc(fin))!=EOF) {
    input.push_back(z);
    occur[z]++;
  }
  int inputsize=z=input.size();
  int escape,endoffile,second;
  if(!pmc1) {
    // check if there are unused symbols. The first one is called escape
    // the second, if available, is called endoffile
    v=0;
    while(v<256 && occur[v])
      v++;
    if(v==256) {
      cout << "ERROR: no available bytes\n";
      return 1;
    }
    escape=v;
    v++;
    while(v<256 && occur[v])
      v++;
    endoffile=v;
    //endoffile=256;
    second=1;
    if (endoffile==256) {
      cout <<"No second byte\n";
      second=0;
    }
    string s2; // LZ-compressed buffer
    int currlen,nextlen=0;
    int currdist,nextdist;
    for(v=0;v<z;v++) { // for each byte of the input
      currlen=nextlen;
      currdist=nextdist;
      if (currlen<18) // if currlen=18 we will not need occurrences
	              // for next step
	findoccur(input,v+1,4095+second,nextlen,nextdist);
      if(currlen && (currlen==18 || currlen>=nextlen)) { // an occurrence,
	                             //not shorter than for the next byte?
	s2+=char(escape); // write the sequence
	s2+=char((currdist-second)%256);
	s2+=char((currdist-second)/256*16+(currlen-3));
	v+=currlen-1;   // we skip several bytes
	findoccur(input,v+1,4095+second,nextlen,nextdist);
      } else
	s2+=input[v];     // otherwise, we just copy the input
    }
    // the end-of-file code depending if we have one free symbol for that
    if (endoffile==256)
      s2+=string(1,escape)+char(0)+char(0); 
    else
      s2+=char(endoffile);
    // now the input is the data compressed with LZ
    input=s2;
    z=input.size();
  }

  // we generate the new occur table
  for(v=0;v<z;v++)
    occur2[input[v]>=0?input[v]:256+input[v]]++;
  // calculate the code for each symbol
  vector<string> code=pmccode(occur2); // code for each symbol
  vector<string> tablecode; // compressed encoding for each element
                                 // of the table
  for(v=0;v<code.size();v++) {
    u=code[v].find_last_of('1');
    if(u==string::npos)
      u=-1; // just for compatibility, because already npos==-1
    tablecode.push_back(string(code[v].size()-u-1,'1')+"0");
  }

  //here we prepare the stream
  string bitstream;
  //first the Huffman tree
  for(v=0;v<256;v++)
    bitstream+=tablecode[v];
  //then the data
  for(v=0;v<z;v++)
    bitstream+=code[input[v]>=0?input[v]:256+input[v]];
  //we complete a byte by adding zeroes
  while(!(bitstream.size()%8))
    bitstream+='0';
  //and now we convert the sequence of bits in a sequence of bytes
  string output;
  for(z=0;z<bitstream.size();z+=8) {
    int a=0;
    for(v=0;v<8;v++)
      a=a*2+(bitstream[z+v]=='1');
    output+=char(a);
  }

  // Output writing
  if(!pmc1) {
    fprintf(fout,"COMPRESCAPE equ %d\n",escape);
    if (endoffile!=256)
      fprintf(fout,"COMPREND equ %d\n",endoffile);
  } else
    fprintf(fout,"uncomprsize equ %d\n",inputsize);
  fprintf(fout,"\n");
  fprintf(fout,"origin:\n");
  cout <<"original size = "<<inputsize<< "      compressed size = "<<output.size()<<"      ratio = "<<double((1000*(inputsize-output.size()))/inputsize)/10<<"%" << endl;
  for(v=0;v<output.size();v++)
    fprintf(fout,"defb %d\n",output[v]>=0?output[v]:256+output[v]);
  fclose(fin);
  fclose(fout);

  exit(0);
}
